import json
from typing import List, Union

import mlx.core as mx
import numpy as np
from mlx_lm.tokenizer_utils import TokenizerWrapper
from outlines.models.transformers import TransformerTokenizer
from outlines.processors.structured import JSONLogitsProcessor

from ..schema import ResponseFormat


class OutlinesLogitsProcessor:
    processed_token_count: int = 0

    def __init__(self, tokenizer: TokenizerWrapper, response_format: ResponseFormat):
        json_schema = response_format.json_schema.schema_def
        self.logits_processor = JSONLogitsProcessor(
            json_schema,
            TransformerTokenizer(tokenizer._tokenizer),
        )

    def _convert_to_numpy_int(
        self, tokens: Union[mx.array, List[int], None]
    ) -> np.ndarray:
        if tokens is None:
            return np.array([], dtype=np.int32)

        if isinstance(tokens, (list, mx.array)):
            tokens_data = tokens.tolist() if isinstance(tokens, mx.array) else tokens
            return np.asarray(tokens_data, dtype=np.int32)

        return np.array([], dtype=np.int32)

    def __call__(self, tokens: mx.array, logits: mx.array) -> mx.array:
        generated_tokens = (
            self._convert_to_numpy_int(tokens[-self.processed_token_count :])
            if self.processed_token_count > 0
            else np.array([], dtype=np.int32)
        )

        logits_data = (
            logits.astype(mx.float32) if logits.dtype == mx.bfloat16 else logits
        )
        logits_flat = logits_data.reshape(-1)

        logits_np = np.asarray(logits_flat.tolist(), dtype=np.float32)
        processed_logits = self.logits_processor(generated_tokens, logits_np)

        self.processed_token_count += 1
        return mx.array(processed_logits).reshape(1, -1)



from fastapi import APIRouter, HTTPException, Request

from .models_service import ModelsService
from .schema import Model, ModelDeletion, ModelList

router = APIRouter(tags=["models"])
models_service = ModelsService()


def extract_model_id_from_path(request: Request) -> str:
    """Extract full model ID from request path"""
    path = request.url.path
    prefix = "/v1/models/" if "/v1/models/" in path else "/models/"
    return path[len(prefix) :]


def handle_model_error(e: Exception) -> None:
    """Handle model-related errors and raise appropriate HTTP exceptions"""
    if isinstance(e, ValueError):
        raise HTTPException(status_code=404, detail=str(e))
    print(f"Error processing request: {str(e)}")
    raise HTTPException(status_code=500, detail=str(e))


@router.get("/models", response_model=ModelList)
@router.get("/v1/models", response_model=ModelList)
async def list_models() -> ModelList:
    """
    Lists the currently available models, and provides basic information about each one
    such as the owner and availability.
    """
    try:
        return models_service.list_models()
    except Exception as e:
        handle_model_error(e)


@router.get("/models/{model_id:path}", response_model=Model)
@router.get("/v1/models/{model_id:path}", response_model=Model)
async def get_model(request: Request) -> Model:
    """
    Retrieves a model instance, providing basic information about the model such as
    the owner and permissioning.
    """
    try:
        model_id = extract_model_id_from_path(request)
        model = models_service.get_model(model_id)
        if model is None:
            raise ValueError(f"Model '{model_id}' not found")
        return model
    except Exception as e:
        handle_model_error(e)


@router.delete("/models/{model_id:path}", response_model=ModelDeletion)
@router.delete("/v1/models/{model_id:path}", response_model=ModelDeletion)
async def delete_model(request: Request) -> ModelDeletion:
    """
    Delete a fine-tuned model from local cache.
    """
    try:
        model_id = extract_model_id_from_path(request)
        return models_service.delete_model(model_id)
    except Exception as e:
        handle_model_error(e)



from fastapi import APIRouter, Depends, HTTPException
from fastapi.responses import JSONResponse, Response
from starlette.responses import PlainTextResponse

from .schema import ResponseFormat, STTRequestForm, TranscriptionResponse
from .whisper_model import STTService

router = APIRouter(tags=["speech-to-text"])


@router.post("/audio/transcriptions", response_model=TranscriptionResponse)
@router.post("/v1/audio/transcriptions", response_model=TranscriptionResponse)
async def create_transcription(request: STTRequestForm = Depends()):
    """
    Transcribe audio file to text.
    """
    stt_service = STTService()
    try:
        result = await stt_service.transcribe(request)

        # Return appropriate response based on format
        if request.response_format == ResponseFormat.TEXT:
            return PlainTextResponse(content=result)
        elif request.response_format in (ResponseFormat.SRT, ResponseFormat.VTT):
            return Response(
                content=result,
                media_type="text/plain",
                headers={
                    "Content-Disposition": f'attachment; filename="transcription.{request.response_format.value.lower()}"'
                },
            )
        else:  # JSON and VERBOSE_JSON
            return JSONResponse(content=result)

    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, Generator, Optional

from .schema import ChatCompletionChunk, ChatCompletionRequest, ChatCompletionResponse


@dataclass
class GenerateResult:
    """Result from generate step"""

    text: str
    token: int
    finish_reason: Optional[str]
    prompt_tokens: int
    generation_tokens: int
    logprobs: Optional[Dict[str, Any]] = None


class BaseTextModel(ABC):
    """Base class for chat models"""

    @abstractmethod
    def generate(
        self,
        request: ChatCompletionRequest,
    ) -> ChatCompletionResponse:
        """Generate completion text with parameters from request"""
        pass

    def stream_generate(
        self,
        request: ChatCompletionRequest,
    ) -> Generator[ChatCompletionChunk, None, None]:
        pass



from typing import Any, Dict, List

from pydantic import BaseModel, Field


class Model(BaseModel):
    """Model information as per OpenAI API specification"""

    id: str = Field(..., description="The model identifier")
    object: str = Field(default="model", description="The object type (always 'model')")
    created: int = Field(
        ..., description="Unix timestamp of when the model was created"
    )
    owned_by: str = Field(..., description="Organization that owns the model")
    config: Dict[str, Any] = Field(
        default_factory=dict, description="Model configuration from config.json"
    )


class ModelList(BaseModel):
    """Response format for list of models"""

    object: str = Field(default="list", description="The object type (always 'list')")
    data: List[Model] = Field(..., description="List of model objects")


class ModelDeletion(BaseModel):
    """Response format for model deletion"""

    id: str = Field(..., description="The ID of the deleted model")
    object: str = Field(default="model", description="The object type (always 'model')")
    deleted: bool = Field(..., description="Whether the model was deleted")



import json
import re
import uuid
from typing import Optional

from ....utils.logger import logger
from ...schema import FunctionCall, ToolCall


def _extract_tools(text: str):
    results = []

    pattern = (
        r'"name"\s*:\s*"([^"]+)"'  # Match name
        r"(?:"  # Start non-capturing group for optional arguments/parameters
        r"[^}]*?"  # Allow any characters in between
        r'(?:"arguments"|"parameters")'  # Match arguments or parameters
        r"\s*:\s*"  # Match colon and whitespace
        r"("  # Start capturing parameter value
        r"\{(?:[^{}]|(?:\{[^{}]*\}))*\}"  # Match nested objects
        r"|\[(?:[^\[\]]|(?:\[[^\[\]]*\]))*\]"  # Match arrays
        r"|null"  # Match null
        r'|"[^"]*"'  # Match strings
        r")"  # End capturing
        r")?"  # Make the entire arguments/parameters section optional
    )

    matches = re.finditer(pattern, text, re.DOTALL)

    matches_list = list(matches)
    for i, match in enumerate(matches_list):
        name, args_str = match.groups()
        # try:
        #     arguments = json.loads(args_str)
        #     print("Successfully parsed arguments as JSON")
        # except json.JSONDecodeError as e:
        #     print(f"Failed to parse JSON: {e}")
        #     arguments = args_str
        results.append({"name": name, "arguments": args_str})

    return results


def parse_tool_calls(text: str) -> Optional[list[ToolCall]]:
    """
    Parse tool calls from text using regex to find JSON patterns containing name and arguments.
    Returns a list of ToolCall objects or None if no valid tool calls are found.
    """
    try:
        tool_calls = _extract_tools(text)
        if tool_calls:
            results = []
            for call in tool_calls:
                # Process arguments
                args = call["arguments"]
                arguments = args if isinstance(args, str) else json.dumps(args)

                tool_call = ToolCall(
                    id=f"call_{uuid.uuid4().hex[:8]}",
                    function=FunctionCall(
                        name=call["name"],
                        arguments=arguments,
                    ),
                )
                results.append(tool_call)
            return results

        return None
    except Exception as e:
        logger.error(f"Error during regex matching: {str(e)}")
        return None



from abc import ABC, abstractmethod
from typing import List, Optional

from mlx_lm.tokenizer_utils import TokenizerWrapper

from ...schema import ChatMessage, Role, Tool, ToolCall, ToolChoice, ToolChoiceType


class ChatTokenizer(ABC):
    """Base class for tools handlers."""

    start_tool_calls: str
    end_tool_calls: str

    def __init__(self, tokenizer: TokenizerWrapper):
        self.tokenizer = tokenizer

    def encode(
        self,
        messages: List[ChatMessage],
        tools: Optional[List[Tool]] = None,
        tool_choice: Optional[ToolChoiceType] = None,
        **kwargs,
    ) -> str:
        """Encode tools and conversation into a prompt string.

        This is a common implementation that uses the tokenizer's chat template.
        Subclasses can override this if they need different behavior.
        """
        schema_tools = None
        if tools:
            schema_tools = [tool.model_dump(exclude_none=True) for tool in tools]

        should_prefill = messages[-1].role == Role.ASSISTANT

        conversation = []
        for message in messages:
            msg_dict = message.model_dump(exclude_none=True)
            if isinstance(msg_dict.get("content"), list):
                msg_dict["content"] = "\n\n".join(
                    item["text"]
                    for item in msg_dict["content"]
                    if item.get("type") == "text"
                )
            conversation.append(msg_dict)

        if should_prefill:
            prompt = self.tokenizer.apply_chat_template(
                conversation=conversation,
                tools=schema_tools,
                tokenize=False,
                continue_final_message=True,
                **kwargs,
            )
        else:
            prompt = self.tokenizer.apply_chat_template(
                conversation=conversation,
                tools=schema_tools,
                tokenize=False,
                add_generation_prompt=True,
                **kwargs,
            )

        if tools:
            if (
                isinstance(tool_choice, ToolChoice)
                and tool_choice == ToolChoice.REQUIRED
            ):
                prompt += self.start_tool_calls

        return prompt

    @abstractmethod
    def decode_stream(self, text: str) -> Optional[List[ToolCall]]:
        """Parse tool calls from model output."""
        pass

    @abstractmethod
    def decode(self, text: str) -> Optional[ChatMessage]:
        """Parse tool calls from model output."""
        pass



from fastapi import APIRouter

from .chat import router as chat_router
from .chat.models import models
from .images import images
from .stt import stt as stt_router
from .tts import tts as tts_router

api_router = APIRouter()
api_router.include_router(stt_router.router)
api_router.include_router(tts_router.router)
api_router.include_router(models.router)
api_router.include_router(images.router)
api_router.include_router(chat_router.router)



from enum import Enum
from typing import List, Optional

from fastapi import File, Form, UploadFile
from pydantic import BaseModel


class ResponseFormat(str, Enum):
    JSON = "json"
    TEXT = "text"
    SRT = "srt"
    VERBOSE_JSON = "verbose_json"
    VTT = "vtt"


class TimestampGranularity(str, Enum):
    WORD = "word"
    SEGMENT = "segment"


class TranscriptionWord(BaseModel):
    word: str
    start: float
    end: float


class Segment(BaseModel):
    id: int
    seek: int
    start: float
    end: float
    text: str
    tokens: List[int]
    temperature: float
    avg_logprob: float
    compression_ratio: float
    no_speech_prob: float


class TranscriptionResponse(BaseModel):
    task: str = "transcribe"
    language: str
    duration: float
    text: str
    words: Optional[List[TranscriptionWord]] = None
    segments: Optional[List[Segment]] = None


class SimpleTranscriptionResponse(BaseModel):
    text: str


class STTRequestForm:
    def __init__(
        self,
        file: UploadFile = File(..., description="The audio file to transcribe"),
        model: str = Form(..., description="ID of the model to use"),
        language: Optional[str] = Form(
            None, description="The language of the input audio in ISO-639-1 format"
        ),
        prompt: Optional[str] = Form(
            None, description="An optional text to guide the model's style"
        ),
        response_format: Optional[ResponseFormat] = Form(
            ResponseFormat.JSON, description="The format of the transcription output"
        ),
        temperature: Optional[float] = Form(
            0.0, description="The sampling temperature, between 0 and 1"
        ),
        timestamp_granularities: Optional[List[str]] = Form(
            default=["segment"],
            alias="timestamp_granularities[]",  # 添加别名以匹配表单字段名
            description="The timestamp granularities to populate (word or segment)",
        ),
    ):
        self.file = file
        self.model = model
        self.language = language
        self.prompt = prompt
        self.response_format = response_format
        self.temperature = temperature

        # 转换 timestamp_granularities 字符串列表为枚举值列表
        self.timestamp_granularities = []
        if timestamp_granularities:
            for gran in timestamp_granularities:
                try:
                    self.timestamp_granularities.append(
                        TimestampGranularity(gran.lower())
                    )
                except ValueError:
                    raise ValueError(
                        f"Invalid timestamp granularity: {gran}. Must be one of: word, segment"
                    )
        else:
            self.timestamp_granularities = [TimestampGranularity.SEGMENT]

        self.validate()

    def validate(self):
        # Validate file extension
        allowed_extensions = {
            "flac",
            "mp3",
            "mp4",
            "mpeg",
            "mpga",
            "m4a",
            "ogg",
            "wav",
            "webm",
        }
        file_extension = self.file.filename.split(".")[-1].lower()
        if file_extension not in allowed_extensions:
            raise ValueError(
                f"File type not supported. Must be one of: {', '.join(allowed_extensions)}"
            )

        # Validate temperature
        if self.temperature is not None and not 0 <= self.temperature <= 1:
            raise ValueError("Temperature must be between 0 and 1")

        # Validate language code if provided
        if self.language and len(self.language) != 2:
            raise ValueError("Language code must be a 2-letter ISO-639-1 code")

        print(f"timestamp_granularities: {self.timestamp_granularities}")

        # 验证 word 时间戳必须使用 verbose_json
        if (
            self.timestamp_granularities
            and TimestampGranularity.WORD in self.timestamp_granularities
            and self.response_format != ResponseFormat.VERBOSE_JSON
        ):
            raise ValueError(
                "Word-level timestamps require response_format=verbose_json"
            )



import logging
import os

# ANSI escape codes for colors
COLORS = {
    "verbose": "\x1b[38;5;244m",  # Grey
    "debug": "\x1b[38;5;33m",  # Blue (#2196F3)
    "info": "\x1b[38;5;77m",  # Green
    "warning": "\x1b[38;5;220m",  # Yellow
    "error": "\x1b[38;5;196m",  # Red
    "critical": "\x1b[38;5;199m",  # Purple
    "reset": "\x1b[0m",
}

# Create logs directory
log_dir = "logs"
if not os.path.exists(log_dir):
    os.makedirs(log_dir)


class ColorFormatter(logging.Formatter):
    """Custom color formatter for logging"""

    format_str = "%(levelname)s: %(message)s"

    FORMATS = {
        logging.DEBUG: COLORS["debug"] + format_str + COLORS["reset"],
        logging.INFO: COLORS["info"] + format_str + COLORS["reset"],
        logging.WARNING: COLORS["warning"] + format_str + COLORS["reset"],
        logging.ERROR: COLORS["error"] + format_str + COLORS["reset"],
        logging.CRITICAL: COLORS["critical"] + format_str + COLORS["reset"],
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def get_logger(name: str = "mlx_omni") -> logging.Logger:
    """Get project logger

    Args:
        name: Module name, typically pass in __name__

    Returns:
        logging.Logger: Configured logger instance
    """
    # Create logger
    _logger = logging.getLogger(name)  # Use module name directly as logger name

    # Set log level
    _logger.setLevel(os.environ.get("MLX_OMNI_LOG_LEVEL", "INFO").upper())

    # If logger already has handlers, it's already configured
    if _logger.handlers:
        return _logger

    # Add console handler
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(ColorFormatter())
    _logger.addHandler(console_handler)

    # Disable log propagation to parent logger
    _logger.propagate = False

    return _logger


# Default logger
logger = get_logger()



import time

from fastapi import APIRouter, HTTPException

from .images_service import ImagesService
from .schema import ImageGenerationRequest, ImageGenerationResponse

router = APIRouter(tags=["images"])


@router.post("/images/generations")
@router.post("/v1/images/generations")
async def create_image(request: ImageGenerationRequest) -> ImageGenerationResponse:
    """
    Creates an image given a prompt.
    """
    try:
        service = ImagesService()

        # Generate images
        images = service.generate_images(request)

        # Create response
        return ImageGenerationResponse(created=int(time.time()), data=images)

    except ValueError as ve:
        raise HTTPException(status_code=400, detail=str(ve))
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))



import os
import tempfile
from pathlib import Path
from typing import Union

from mlx_whisper import transcribe
from mlx_whisper.writers import WriteSRT, WriteVTT

from .schema import (
    ResponseFormat,
    STTRequestForm,
    TranscriptionResponse,
    TranscriptionWord,
)


class WhisperModel:

    async def _save_upload_file(self, file) -> str:
        suffix = Path(file.filename).suffix
        with tempfile.NamedTemporaryFile(delete=False, suffix=suffix) as tmp:
            content = await file.read()
            tmp.write(content)
            return tmp.name

    def generate(self, audio_path: str, request: STTRequestForm):
        word_timestamps = False
        if request.timestamp_granularities:
            word_timestamps = "word" in [
                g.value for g in request.timestamp_granularities
            ]

        print(f"word_timestamps: {word_timestamps}")
        result = transcribe(
            audio=audio_path,
            path_or_hf_repo=request.model,
            temperature=request.temperature,
            initial_prompt=request.prompt,
            language=request.language,
            word_timestamps=word_timestamps,
            verbose=False,
            condition_on_previous_text=True,
        )
        return result

    def _generate_subtitle_file(self, result: dict, format: str) -> str:
        temp_dir = None
        temp_file = None
        try:
            temp_dir = tempfile.mkdtemp()
            temp_file = os.path.join(temp_dir, f"temp.{format}")

            if format == "srt":
                writer = WriteSRT(temp_dir)
            else:  # vtt
                writer = WriteVTT(temp_dir)

            writer(result, temp_file)

            with open(temp_file, "r", encoding="utf-8") as f:
                return f.read()

        finally:
            if temp_file and os.path.exists(temp_file):
                os.unlink(temp_file)
            if temp_dir and os.path.exists(temp_dir):
                os.rmdir(temp_dir)

    def _format_response(
        self, result: dict, request: STTRequestForm
    ) -> Union[dict, str, TranscriptionResponse]:
        if request.response_format == ResponseFormat.TEXT:
            return result["text"]

        elif request.response_format == ResponseFormat.SRT:
            return self._generate_subtitle_file(result, "srt")

        elif request.response_format == ResponseFormat.VTT:
            return self._generate_subtitle_file(result, "vtt")

        elif request.response_format == ResponseFormat.VERBOSE_JSON:
            return result

        elif request.response_format == ResponseFormat.JSON:
            return {"text": result["text"]}

        else:
            text = result.get("text", "")
            language = result.get("language", "en")

            duration = 0
            if "segments" in result:
                for segment in result["segments"]:
                    if "end" in segment:
                        duration = max(duration, segment["end"])

            words = []
            if request.timestamp_granularities and "word" in [
                g.value for g in request.timestamp_granularities
            ]:
                for segment in result.get("segments", []):
                    for word_data in segment.get("words", []):
                        word = TranscriptionWord(
                            word=word_data["word"],
                            start=word_data["start"],
                            end=word_data["end"],
                        )
                        words.append(word)

            return TranscriptionResponse(
                task="transcribe",
                language=language,
                duration=duration,
                text=text,
                words=words if words else None,
            )


class STTService:
    def __init__(self):
        self.model = WhisperModel()

    async def transcribe(
        self,
        request: STTRequestForm,
    ) -> Union[dict, str, TranscriptionResponse]:
        try:
            audio_path = await self.model._save_upload_file(request.file)
            result = self.model.generate(audio_path=audio_path, request=request)
            response = self.model._format_response(result, request)
            Path(audio_path).unlink(missing_ok=True)
            return response

        except Exception as e:
            if "audio_path" in locals():
                Path(audio_path).unlink(missing_ok=True)
            raise e



import json
import uuid
from typing import List, Optional

from mlx_lm.tokenizer_utils import TokenizerWrapper

from ....utils.logger import logger
from ...schema import (
    ChatMessage,
    FunctionCall,
    Role,
    SpecificToolChoice,
    Tool,
    ToolCall,
    ToolChoiceType,
)
from .chat_tokenizer import ChatTokenizer
from .utils import parse_tool_calls


class HuggingFaceChatTokenizer(ChatTokenizer):
    """Tools handler for Llama models.
    https://huggingface.co/blog/unified-tool-use
    """

    def __init__(self, tokenizer: TokenizerWrapper):
        super().__init__(tokenizer)
        self.start_tool_calls = "<tool_call>\n"
        self.end_tool_calls = "</tool_call>"
        self.strict_mode = False
        self.pre_fill_tools_prompt = ""

    def encode(
        self,
        messages: List[ChatMessage],
        tools: Optional[List[Tool]] = None,
        tool_choice: Optional[ToolChoiceType] = None,
        **kwargs,
    ):
        prompt = super().encode(messages, tools, tool_choice, **kwargs)

        if tools:
            if isinstance(tool_choice, SpecificToolChoice):
                self.pre_fill_tools_prompt += self.start_tool_calls
                function_name = tool_choice.function["name"]

                self.pre_fill_tools_prompt += (
                    f"""{{"name": "{function_name}", "arguments":"""
                )

        return prompt + self.pre_fill_tools_prompt

    def decode_stream(self, text: str, delta_text: str) -> Optional[List[ToolCall]]:
        pass

    def _parse_strict_tools(self, text: str) -> Optional[List[ToolCall]]:
        tool_calls = []
        logger.debug(f"_parse_strict_tools: {text}")

        if (
            text.strip().startswith(self.start_tool_calls)
            and self.end_tool_calls in text
        ):
            try:
                # Remove tool call tags and parse JSON directly
                json_str = text[
                    len(self.start_tool_calls) : text.find(self.end_tool_calls)
                ].strip()
                tool_data = json.loads(json_str)

                if isinstance(tool_data, dict) and "name" in tool_data:
                    # Get arguments and ensure they're a JSON string
                    args = tool_data.get("arguments", tool_data.get("parameters", {}))
                    if isinstance(args, str):
                        # Already a JSON string
                        arguments = args
                    else:
                        # Convert dict to JSON string
                        arguments = json.dumps(args)

                    tool_calls.append(
                        ToolCall(
                            id=f"call_{uuid.uuid4().hex[:8]}",
                            function=FunctionCall(
                                name=tool_data["name"],
                                arguments=arguments,
                            ),
                        )
                    )
            except (json.JSONDecodeError, KeyError, ValueError) as e:
                logger.error(f"Error parsing tool call: {e}")
                return None

        return tool_calls if tool_calls else None

    def decode(self, text: str) -> Optional[ChatMessage]:
        """Parse tool calls from model output."""
        response = self.pre_fill_tools_prompt + text

        if self.strict_mode:
            tool_calls = self._parse_strict_tools(response)
        else:
            tool_calls = parse_tool_calls(response)

        return ChatMessage(
            role=Role.ASSISTANT,
            content=None if tool_calls else text,
            tool_calls=tool_calls,
        )



import json
from typing import Generator

from fastapi import APIRouter
from fastapi.responses import JSONResponse, StreamingResponse

from .mlx.models import load_model
from .schema import ChatCompletionRequest, ChatCompletionResponse
from .text_models import BaseTextModel

router = APIRouter(tags=["chat—completions"])


@router.post("/chat/completions", response_model=ChatCompletionResponse)
@router.post("/v1/chat/completions", response_model=ChatCompletionResponse)
async def create_chat_completion(request: ChatCompletionRequest):
    """Create a chat completion"""

    text_model = _create_text_model(
        request.model, request.get_extra_params().get("adapter_path")
    )

    if not request.stream:
        completion = text_model.generate(request)
        return JSONResponse(content=completion.model_dump(exclude_none=True))

    async def event_generator() -> Generator[str, None, None]:
        for chunk in text_model.stream_generate(request):
            yield f"data: {json.dumps(chunk.model_dump(exclude_none=True))}\n\n"

        yield "data: [DONE]\n\n"

    return StreamingResponse(
        event_generator(),
        media_type="text/event-stream",
        headers={
            "Cache-Control": "no-cache",
            "Connection": "keep-alive",
        },
    )


_last_model_id = None
_last_text_model = None


def _create_text_model(model_id: str, adapter_path: str = None) -> BaseTextModel:
    global _last_model_id, _last_text_model
    if model_id == _last_model_id:
        return _last_text_model

    model = load_model(model_id, adapter_path)
    _last_text_model = model
    _last_model_id = model_id
    return model



import importlib
import json
import logging
from typing import Dict, List, Optional, Tuple, Type

from huggingface_hub import CachedRepoInfo, scan_cache_dir

from .schema import Model, ModelDeletion, ModelList

MODEL_REMAPPING = {
    "mistral": "llama",  # mistral is compatible with llama
    "phi-msft": "phixtral",
    "falcon_mamba": "mamba",
}


class ModelCacheScanner:
    """Scanner for finding and managing mlx-lm compatible models in the local cache."""

    def __init__(self):
        self._cache_info = None

    @property
    def cache_info(self):
        """Lazy load and cache the scan_cache_dir result"""
        if self._cache_info is None:
            self._cache_info = scan_cache_dir()
        return self._cache_info

    def _refresh_cache_info(self):
        """Force refresh the cache info"""
        self._cache_info = scan_cache_dir()

    def _get_model_classes(self, config: dict) -> Optional[Tuple[Type, Type]]:
        """
        Try to retrieve the model and model args classes based on the configuration.
        https://github.com/ml-explore/mlx-examples/blob/1e0766018494c46bc6078769278b8e2a360503dc/llms/mlx_lm/utils.py#L81

        Args:
            config (dict): The model configuration

        Returns:
            Optional tuple of (Model class, ModelArgs class) if model type is supported
        """
        try:
            model_type = config.get("model_type")
            model_type = MODEL_REMAPPING.get(model_type, model_type)
            if not model_type:
                return None

            # Try to import the model architecture module
            arch = importlib.import_module(f"mlx_lm.models.{model_type}")
            return arch.Model, arch.ModelArgs

        except ImportError:
            logging.debug(f"Model type {model_type} not supported by mlx-lm")
            return None
        except Exception as e:
            logging.warning(f"Error checking model compatibility: {str(e)}")
            return None

    def is_model_supported(self, config_data: Dict) -> bool:
        return self._get_model_classes(config_data) is not None

    def find_models_in_cache(self) -> List[Tuple[CachedRepoInfo, Dict]]:
        """
        Scan local cache for available models that are compatible with mlx-lm.

        Returns:
            List of tuples containing (CachedRepoInfo, config_dict)
        """
        supported_models = []

        for repo_info in self.cache_info.repos:
            if repo_info.repo_type != "model":
                continue

            first_revision = next(iter(repo_info.revisions), None)
            if not first_revision:
                continue

            config_file = next(
                (f for f in first_revision.files if f.file_name == "config.json"), None
            )
            if not config_file:
                continue

            try:
                with open(config_file.file_path, "r") as f:
                    config_data = json.load(f)
                if self.is_model_supported(config_data):
                    supported_models.append((repo_info, config_data))
            except Exception as e:
                logging.error(
                    f"Error reading config.json for {repo_info.repo_id}: {str(e)}"
                )

        return supported_models

    def get_model_info(self, model_id: str) -> Optional[Tuple[CachedRepoInfo, Dict]]:
        for repo_info in self.cache_info.repos:
            if repo_info.repo_id == model_id and repo_info.repo_type == "model":
                first_revision = next(iter(repo_info.revisions), None)
                if not first_revision:
                    continue

                config_file = next(
                    (f for f in first_revision.files if f.file_name == "config.json"),
                    None,
                )
                if not config_file:
                    continue

                try:
                    with open(config_file.file_path, "r") as f:
                        config_data = json.load(f)
                    if self.is_model_supported(config_data):
                        return (repo_info, config_data)
                    else:
                        logging.warning(
                            f"Model {model_id} found but not compatible with mlx-lm"
                        )
                except Exception as e:
                    logging.error(
                        f"Error reading config.json for {repo_info.repo_id}: {str(e)}"
                    )

        return None

    def delete_model(self, model_id: str) -> bool:
        for repo_info in self.cache_info.repos:
            if repo_info.repo_id == model_id:
                revision_hashes = [rev.commit_hash for rev in repo_info.revisions]
                if not revision_hashes:
                    return False

                try:
                    delete_strategy = self.cache_info.delete_revisions(*revision_hashes)
                    logging.info(
                        f"Model '{model_id}': Will free {delete_strategy.expected_freed_size_str}"
                    )
                    delete_strategy.execute()
                    logging.info(f"Model '{model_id}': Cache deletion completed")
                    self._refresh_cache_info()
                    return True
                except Exception as e:
                    logging.error(f"Error deleting model '{model_id}': {str(e)}")
                    raise

        return False


class ModelsService:
    def __init__(self):
        self.scanner = ModelCacheScanner()
        self.available_models = self._scan_models()

    def _scan_models(self) -> List[Tuple[CachedRepoInfo, Dict]]:
        """Scan local cache for available CausalLM models"""
        try:
            return self.scanner.find_models_in_cache()
        except Exception as e:
            print(f"Error scanning cache: {str(e)}")
            return []

    @staticmethod
    def _get_model_owner(model_id: str) -> str:
        """Extract owner from model ID (part before the /)"""
        return model_id.split("/")[0] if "/" in model_id else model_id

    def list_models(self) -> ModelList:
        """List all available models"""
        models = []
        for repo_info, config_data in self.available_models:
            models.append(
                Model(
                    id=repo_info.repo_id,
                    created=int(repo_info.last_modified),
                    owned_by=self._get_model_owner(repo_info.repo_id),
                    config=config_data,
                )
            )
        return ModelList(data=models)

    def get_model(self, model_id: str) -> Optional[Model]:
        """Get information about a specific model"""
        model_info = self.scanner.get_model_info(model_id)
        if model_info:
            repo_info, config_data = model_info
            return Model(
                id=model_id,
                created=int(repo_info.last_modified),
                owned_by=self._get_model_owner(model_id),
                config=config_data,
            )
        return None

    def delete_model(self, model_id: str) -> ModelDeletion:
        """Delete a model from local cache"""
        if not self.scanner.delete_model(model_id):
            raise ValueError(f"Model '{model_id}' not found in cache")

        self.available_models = self._scan_models()
        return ModelDeletion(id=model_id, deleted=True)



from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field, field_validator


class ImageSize(str, Enum):
    S256x256 = "256x256"
    S512x512 = "512x512"
    S1024x1024 = "1024x1024"
    S1792x1024 = "1792x1024"
    S1024x1792 = "1024x1792"


class ImageQuality(str, Enum):
    STANDARD = "standard"
    HD = "hd"


class ImageStyle(str, Enum):
    VIVID = "vivid"
    NATURAL = "natural"


class ResponseFormat(str, Enum):
    URL = "url"
    B64_JSON = "b64_json"


class ImageGenerationRequest(BaseModel):
    prompt: str = Field(..., max_length=4000)
    model: Optional[str] = Field(
        default="argmaxinc/mlx-FLUX.1-schnell",
        description="The model to use for image generation",
    )
    n: Optional[int] = Field(default=1, ge=1, le=10)
    quality: Optional[ImageQuality] = Field(default=ImageQuality.STANDARD)
    response_format: Optional[ResponseFormat] = Field(default=ResponseFormat.B64_JSON)
    size: Optional[ImageSize] = Field(default=ImageSize.S1024x1024)
    style: Optional[ImageStyle] = Field(default=ImageStyle.VIVID)
    user: Optional[str] = None

    # Allow any additional fields
    class Config:
        extra = "allow"  # This allows additional fields not defined in the model

    def get_extra_params(self) -> Dict[str, Any]:
        """Get all extra parameters that aren't part of the standard OpenAI API."""
        standard_fields = {
            "prompt",
            "model",
            "n",
            "quality",
            "response_format",
            "size",
            "style",
            "user",
        }
        return {k: v for k, v in self.model_dump().items() if k not in standard_fields}

    @field_validator("prompt")
    def validate_prompt_length(cls, v, values):
        max_length = 4000
        if len(v) > max_length:
            raise ValueError(
                f"Prompt length exceeds maximum of {max_length} characters"
            )
        return v


class ImageObject(BaseModel):
    url: Optional[str] = None
    b64_json: Optional[str] = None
    revised_prompt: Optional[str] = None


class ImageGenerationResponse(BaseModel):
    created: int
    data: List[ImageObject]



import argparse
import os

import uvicorn
from fastapi import FastAPI

from .middleware.logging import RequestResponseLoggingMiddleware
from .routers import api_router

app = FastAPI(title="MLX Omni Server")

# Add request/response logging middleware with custom levels
app.add_middleware(
    RequestResponseLoggingMiddleware,
    # exclude_paths=["/health"]
)

app.include_router(api_router)


def build_parser():
    """Create and configure the argument parser for the server."""
    parser = argparse.ArgumentParser(description="MLX Omni Server")
    parser.add_argument(
        "--host",
        type=str,
        default="0.0.0.0",
        help="Host to bind the server to, defaults to 0.0.0.0",
    )
    parser.add_argument(
        "--port",
        type=int,
        default=10240,
        help="Port to bind the server to, defaults to 10240",
    )
    parser.add_argument(
        "--workers",
        type=int,
        default=1,
        help="Number of workers to use, defaults to 1",
    )
    parser.add_argument(
        "--log-level",
        type=str,
        default="info",
        choices=["debug", "info", "warning", "error", "critical"],
        help="Set the logging level, defaults to info",
    )
    return parser


def start():
    """Start the MLX Omni Server."""
    parser = build_parser()
    args = parser.parse_args()

    # Set log level through environment variable
    os.environ["MLX_OMNI_LOG_LEVEL"] = args.log_level

    # Start server with uvicorn
    uvicorn.run(
        "mlx_omni_server.main:app",
        host=args.host,
        port=args.port,
        log_level=args.log_level,
        use_colors=True,
        workers=args.workers,
    )



from pathlib import Path

from f5_tts_mlx.generate import generate

from .schema import TTSRequest


class F5Model:

    def __init__(self):
        pass

    def generate_audio(self, request: TTSRequest, output_path):
        generate(
            model_name=request.model,
            generation_text=request.input,
            speed=request.speed,
            output_path=output_path,
            **(request.get_extra_params() or {}),
        )


class TTSService:
    model: F5Model

    def __init__(self):
        self.model = F5Model()
        self.sample_audio_path = Path("sample.wav")

    async def generate_speech(
        self,
        request: TTSRequest,
    ) -> bytes:
        try:
            self.model.generate_audio(
                request=request, output_path=self.sample_audio_path
            )
            with open(self.sample_audio_path, "rb") as audio_file:
                audio_content = audio_file.read()
            self.sample_audio_path.unlink(missing_ok=True)
            return audio_content
        except Exception as e:
            raise Exception(f"Error reading audio file: {str(e)}")



import json
import uuid
from typing import List, Optional

from mlx_lm.tokenizer_utils import TokenizerWrapper

from ...schema import ChatMessage, FunctionCall, Role, ToolCall
from .chat_tokenizer import ChatTokenizer


class MistralChatTokenizer(ChatTokenizer):
    """Tools handler for Llama models."""

    def __init__(self, tokenizer: TokenizerWrapper):
        super().__init__(tokenizer)
        self.start_tool_calls = "[TOOL_CALLS]"
        self.end_tool_calls = ""

    def decode_stream(self, text: str, delta_text: str) -> Optional[List[ToolCall]]:
        pass

    def decode(self, text: str) -> Optional[ChatMessage]:
        """Parse tool calls from model output.

        The model outputs function calls in the format:
        [TOOL_CALLS] [{"name": "get_current_weather", "arguments": {"location": "Boston, MA"}},
                     {"name": "get_current_weather", "arguments": {"location": "Boston, MA"}}]

        Args:
            text: The model output text containing tool calls

        Returns:
            ChatMessage: A message containing the parsed tool calls
        """
        # Look for JSON patterns in the text
        tool_calls = []

        if text.startswith(self.start_tool_calls):
            try:
                # Extract the JSON array from between square brackets after [TOOL_CALLS]
                json_str = text[len(self.start_tool_calls) :].strip()
                if json_str.startswith("[") and json_str.endswith("]"):
                    json_str = json_str.strip("[]").strip()
                    # Try to parse as array first
                    try:
                        tool_data = json.loads(f"[{json_str}]")
                    except json.JSONDecodeError:
                        # If array parsing fails, try single object
                        tool_data = json.loads(json_str)

                    # Handle both single object and array of objects
                    if isinstance(tool_data, dict):
                        tool_data = [tool_data]
                    elif not isinstance(tool_data, list):
                        raise ValueError(
                            "Invalid tool call format: expected dict or list"
                        )

                    for call in tool_data:
                        if not isinstance(call, dict) or "name" not in call:
                            continue

                        # Get arguments and ensure they're a JSON string
                        args = call.get("arguments", call.get("parameters", {}))
                        if isinstance(args, str):
                            # Already a JSON string
                            arguments = args
                        else:
                            # Convert dict to JSON string
                            arguments = json.dumps(args)

                        tool_calls.append(
                            ToolCall(
                                id=f"call_{uuid.uuid4().hex[:8]}",
                                function=FunctionCall(
                                    name=call["name"],
                                    arguments=arguments,
                                ),
                            )
                        )
                else:
                    # Invalid format, return original text
                    return ChatMessage(
                        role=Role.ASSISTANT,
                        content=text,
                        tool_calls=None,
                    )
            except (json.JSONDecodeError, KeyError, ValueError) as e:
                print(f"Error parsing tool call: {e}")
                return ChatMessage(
                    role=Role.ASSISTANT,
                    content=text,
                    tool_calls=None,
                )

        return ChatMessage(
            role=Role.ASSISTANT,
            content=None if tool_calls else text,
            tool_calls=tool_calls if tool_calls else None,
        )



from enum import Enum
from typing import Any, Dict, Optional

from pydantic import BaseModel, Field, field_validator


class AudioFormat(str, Enum):
    MP3 = "mp3"
    OPUS = "opus"
    AAC = "aac"
    FLAC = "flac"
    WAV = "wav"
    PCM = "pcm"


class TTSRequest(BaseModel):
    model: str = Field(..., description="TTS model to use")
    input: str = Field(..., max_length=4096)
    voice: str = Field(
        ..., description="Voice to use (e.g. alloy, echo, fable, onyx, nova, shimmer)"
    )
    response_format: Optional[AudioFormat] = Field(default=AudioFormat.WAV)
    speed: Optional[float] = Field(default=1.0)

    # Allow any additional fields
    class Config:
        extra = "allow"  # This allows additional fields not defined in the model

    def get_extra_params(self) -> Dict[str, Any]:
        """Get all extra parameters that aren't part of the standard OpenAI API."""
        standard_fields = {"model", "input", "voice", "response_format", "speed"}
        return {k: v for k, v in self.model_dump().items() if k not in standard_fields}

    @field_validator("speed")
    def validate_speed(cls, v):
        if v < 0.25 or v > 4.0:
            raise ValueError("Speed must be between 0.25 and 4.0")
        return v

    @field_validator("model")
    def validate_model(cls, v):
        valid_models = ["lucasnewman/f5-tts-mlx"]
        if v not in valid_models:
            raise ValueError(f'Model must be one of: {", ".join(valid_models)}')
        return v



from typing import List, NamedTuple, Optional, Union

from mlx_lm.tokenizer_utils import TokenizerWrapper


class StopCondition(NamedTuple):
    stop_met: bool
    trim_length: int


class StopTokensChecker:
    """A class to handle custom stop sequence checks."""

    def __init__(
        self,
        stop_words: Optional[Union[str, List[str]]],
        tokenizer: TokenizerWrapper,
    ):
        """Initialize the TokenStopChecker.

        Args:
            stop_words: Words or phrases that should trigger stopping
            tokenizer: The tokenizer to use for encoding stop words
        """
        self._stop_id_sequences = self._prepare_stop_sequences(stop_words, tokenizer)

    def _prepare_stop_sequences(
        self, stop_words: Optional[Union[str, List[str]]], tokenizer: TokenizerWrapper
    ) -> List[List[int]]:
        """Prepare stop sequences by converting words to token IDs.

        Args:
            stop_words: Words to convert to token sequences
            tokenizer: Tokenizer to use for encoding

        Returns:
            List of token ID sequences
        """
        if not stop_words:
            return []

        words = [stop_words] if isinstance(stop_words, str) else stop_words
        return [
            tokenizer.encode(word, add_special_tokens=False)
            for word in words
            if word  # Skip empty strings
        ]

    def check_stop_condition(self, tokens: List[int]) -> StopCondition:
        """Check if the token sequence meets any custom stop criteria.

        Args:
            tokens: The current sequence of generated tokens

        Returns:
            StopCondition indicating if/how generation should stop
        """
        if not tokens or not self._stop_id_sequences:
            return StopCondition(stop_met=False, trim_length=0)

        # Check stop sequences
        for stop_ids in self._stop_id_sequences:
            stop_len = len(stop_ids)
            if len(tokens) >= stop_len and tokens[-stop_len:] == stop_ids:
                # Look for partial matches to trim
                prefix_len = self._find_prefix_length(tokens[:-stop_len], stop_ids[0])
                return StopCondition(stop_met=True, trim_length=stop_len + prefix_len)

        return StopCondition(stop_met=False, trim_length=0)

    @staticmethod
    def _find_prefix_length(tokens: List[int], first_stop_token: int) -> int:
        """Find length of matching prefix tokens.

        Args:
            tokens: Token sequence to check
            first_stop_token: First token of stop sequence

        Returns:
            Length of matching prefix
        """
        prefix_len = 0
        for i in range(len(tokens) - 1, -1, -1):
            if tokens[i] == first_stop_token:
                prefix_len += 1
            else:
                break
        return prefix_len



import time
import uuid
from typing import Any, Dict, Generator, Optional

import mlx.core as mx
from mlx_lm.sample_utils import make_logits_processors, make_sampler
from mlx_lm.tokenizer_utils import TokenizerWrapper
from mlx_lm.utils import GenerationResponse, stream_generate

from ...utils.logger import logger
from ..schema import (
    ChatCompletionChoice,
    ChatCompletionChunk,
    ChatCompletionChunkChoice,
    ChatCompletionRequest,
    ChatCompletionResponse,
    ChatCompletionUsage,
    ChatMessage,
    Role,
)
from ..text_models import BaseTextModel, GenerateResult
from .outlines_logits_processor import OutlinesLogitsProcessor
from .stop_tokens_checker import StopTokensChecker
from .tools.chat_tokenizer import ChatTokenizer


class MLXModel(BaseTextModel):
    """MLX Chat Model wrapper with internal parameter management"""

    def __init__(self, model_id: str, model, tokenizer: ChatTokenizer):
        self._model_id = model_id
        self._model = model
        self._default_max_tokens = 2048
        self._default_temperature = 1.0
        self._default_top_p = 1.0
        self._default_top_k = -1
        self._chat_tokenizer = tokenizer
        logger.info(f"Initialized MLXModel with model_id: {model_id}")

    def _get_generation_params(self, request: ChatCompletionRequest) -> Dict[str, Any]:
        params = request.get_extra_params()
        known_params = {
            "top_k",
            "min_tokens_to_keep",
            "min_p",
            "adapter_path",
        }
        return {k: v for k, v in params.items() if k not in known_params}

    def _process_logprobs(
        self,
        tokenizer: TokenizerWrapper,
        response: GenerationResponse,
        top_k: Optional[int],
    ) -> Optional[Dict[str, Any]]:
        """Process logprobs information from generation response to match OpenAI format"""
        current_token = response.token
        current_logprobs = response.logprobs

        # Get current token info
        token_str = tokenizer.decode([current_token])
        token_logprob = current_logprobs[current_token].item()
        token_bytes = token_str.encode("utf-8")

        # Base token info
        token_info = {
            "token": token_str,
            "logprob": token_logprob,
            "bytes": list(token_bytes),
        }

        # Process top logprobs
        top_logprobs = []
        if top_k is not None:
            # Get indices of top_k tokens
            top_indices = mx.argpartition(-current_logprobs, kth=top_k - 1)[:top_k]
            top_probs = current_logprobs[top_indices]

            # Create detailed token information for each top token
            for idx, logprob in zip(top_indices.tolist(), top_probs.tolist()):
                token = tokenizer.decode([idx])
                token_bytes = token.encode("utf-8")
                top_logprobs.append(
                    {"token": token, "logprob": logprob, "bytes": list(token_bytes)}
                )

        return {**token_info, "top_logprobs": top_logprobs}

    def _stream_generate(
        self,
        prompt: str,
        request: ChatCompletionRequest,
    ) -> Generator[GenerationResponse, None, None]:
        try:
            params = self._get_generation_params(request)

            tokenizer = self._chat_tokenizer.tokenizer
            stop_checker = None
            if request.stop:
                stop_checker = StopTokensChecker(
                    stop_words=request.stop,
                    tokenizer=tokenizer,
                )

            logits_processors = None
            if request.response_format and request.response_format.json_schema:
                logits_processors = [
                    OutlinesLogitsProcessor(
                        self._chat_tokenizer.tokenizer, request.response_format
                    )
                ]
            else:
                if request.presence_penalty:
                    logits_processors = make_logits_processors(
                        repetition_penalty=request.presence_penalty
                    )

            current_tokens = []
            last_text = ""

            max_completion_tokens = (
                request.max_completion_tokens
                or request.max_tokens
                or self._default_max_tokens
            )
            sampler = make_sampler(
                temp=request.temperature or self._default_temperature,
                top_p=request.top_p or self._default_top_p,
                min_p=params.get("min_p", 0.0),
                min_tokens_to_keep=params.get("min_tokens_to_keep", 1),
                top_k=params.get("top_k", self._default_top_k),
            )

            for response in stream_generate(
                model=self._model,
                tokenizer=tokenizer,
                prompt=prompt,
                max_tokens=max_completion_tokens,
                sampler=sampler,
                logits_processors=logits_processors,
                **params,
            ):
                if response.finish_reason is not None:
                    break

                current_tokens.append(response.token)

                logprobs = None
                if request.logprobs:
                    logprobs = self._process_logprobs(
                        tokenizer, response, request.top_logprobs
                    )

                finish_reason = response.finish_reason
                should_trim = False
                if request.stop and stop_checker:
                    stop_condition = stop_checker.check_stop_condition(current_tokens)
                    if stop_condition.stop_met:
                        finish_reason = "stop"
                        if stop_condition.trim_length > 0:
                            current_tokens = current_tokens[
                                : -stop_condition.trim_length
                            ]
                            should_trim = True

                text = tokenizer.decode(current_tokens)
                delta_text = text[len(last_text) :]

                if delta_text or should_trim:
                    yield GenerateResult(
                        text=delta_text,
                        token=response.token,
                        finish_reason=finish_reason,
                        prompt_tokens=response.prompt_tokens,
                        generation_tokens=response.generation_tokens,
                        logprobs=logprobs,
                    )
                    last_text = text

                if should_trim:
                    break

        except Exception as e:
            logger.error(f"Error during stream generation: {str(e)}", exc_info=True)
            raise

    def generate(
        self,
        request: ChatCompletionRequest,
    ) -> ChatCompletionResponse:
        try:
            completion = ""
            logprobs_result_list = []
            current_tokens = []
            finish_reason = "stop"
            result = None

            prompt = self._chat_tokenizer.encode(
                messages=request.messages,
                tools=request.tools,
                tool_choice=request.tool_choice if request.tool_choice else None,
            )
            logger.debug(f"Encoded prompt:\n{prompt}")

            for result in self._stream_generate(
                prompt=prompt,
                request=request,
            ):
                current_tokens.append(result.token)
                completion = self._chat_tokenizer.tokenizer.decode(current_tokens)

                if request.logprobs:
                    logprobs_result_list.append(result.logprobs)

                if not prompt:
                    prompt = result.text

                if result.finish_reason:
                    finish_reason = result.finish_reason

            if result is None:
                raise RuntimeError("No tokens generated")

            logger.debug(f"Model Response:\n{completion}")
            if request.tools:
                message = self._chat_tokenizer.decode(completion)
            else:
                message = ChatMessage(role=Role.ASSISTANT, content=completion)

            return ChatCompletionResponse(
                id=f"chatcmpl-{uuid.uuid4().hex[:10]}",
                created=int(time.time()),
                model=request.model,
                choices=[
                    ChatCompletionChoice(
                        index=0,
                        message=message,
                        finish_reason=(
                            "tool_calls" if message.tool_calls else finish_reason
                        ),
                        logprobs=(
                            {"content": logprobs_result_list}
                            if logprobs_result_list
                            else None
                        ),
                    )
                ],
                usage=ChatCompletionUsage(
                    prompt_tokens=result.prompt_tokens,
                    completion_tokens=result.generation_tokens,
                    total_tokens=result.prompt_tokens + result.generation_tokens,
                ),
            )
        except Exception as e:
            logger.error(f"Failed to generate completion: {str(e)}", exc_info=True)
            raise RuntimeError(f"Failed to generate completion: {str(e)}")

    def stream_generate(
        self,
        request: ChatCompletionRequest,
    ) -> Generator[ChatCompletionChunk, None, None]:
        try:
            chat_id = f"chatcmpl-{uuid.uuid4().hex[:10]}"

            prompt = self._chat_tokenizer.encode(
                messages=request.messages,
                tools=request.tools,
            )
            logger.debug(f"Encoded prompt:\n{prompt}")

            completion = ""
            for result in self._stream_generate(
                prompt=prompt,
                request=request,
            ):
                created = int(time.time())
                completion += result.text
                yield ChatCompletionChunk(
                    id=chat_id,
                    created=created,
                    model=request.model,
                    choices=[
                        ChatCompletionChunkChoice(
                            index=0,
                            delta=ChatMessage(role=Role.ASSISTANT, content=result.text),
                            finish_reason=result.finish_reason,
                            logprobs=result.logprobs,
                        )
                    ],
                )

            if request.stream_options and request.stream_options.include_usage:
                created = int(time.time())
                yield ChatCompletionChunk(
                    id=chat_id,
                    created=created,
                    model=request.model,
                    choices=[
                        ChatCompletionChunkChoice(
                            index=0,
                            delta=ChatMessage(role=Role.ASSISTANT),
                            finish_reason=None,
                            logprobs=None,
                        )
                    ],
                    usage=ChatCompletionUsage(
                        prompt_tokens=result.prompt_tokens,
                        completion_tokens=result.generation_tokens,
                        total_tokens=result.prompt_tokens + result.generation_tokens,
                    ),
                )

        except Exception as e:
            logger.error(f"Error during stream generation: {str(e)}", exc_info=True)
            raise



from typing import Type

from mlx_lm.tokenizer_utils import TokenizerWrapper
from mlx_lm.utils import get_model_path, load, load_config

from ..text_models import BaseTextModel
from .mlx_model import MLXModel
from .tools.chat_tokenizer import ChatTokenizer
from .tools.hugging_face import HuggingFaceChatTokenizer
from .tools.llama3 import Llama3ChatTokenizer
from .tools.mistral import MistralChatTokenizer


def load_tools_handler(model_type: str, tokenizer: TokenizerWrapper) -> ChatTokenizer:
    """Factory function to load appropriate tools handler based on model ID."""
    handlers: dict[str, Type[ChatTokenizer]] = {
        # Llama models
        "llama": Llama3ChatTokenizer,
        "mistral": MistralChatTokenizer,
        "qwen2": HuggingFaceChatTokenizer,
    }

    # Get handler class based on model ID or use Llama handler as default
    handler_class = handlers.get(model_type, HuggingFaceChatTokenizer)
    return handler_class(tokenizer)


def load_model(model_id: str, adapter_path: str = None) -> BaseTextModel:
    """Load a model and tokenizer from the given model ID."""
    model, tokenizer = load(
        model_id,
        tokenizer_config={"trust_remote_code": True},
        adapter_path=adapter_path,
    )

    model_path = get_model_path(model_id)
    config = load_config(model_path)

    chat_tokenizer = load_tools_handler(config["model_type"], tokenizer)

    return MLXModel(model_id=model_id, model=model, tokenizer=chat_tokenizer)



import json
import re
from enum import Enum
from typing import Any, Dict, List, Optional, Set, Union

from pydantic import BaseModel, Field, field_validator


class ToolType(str, Enum):
    FUNCTION = "function"


class FunctionParameters(BaseModel):
    type: Optional[str] = None
    properties: Optional[Dict[str, Any]] = None
    required: Optional[List[str]] = None


class Function(BaseModel):
    name: str = Field(..., max_length=64, pattern=r"^[a-zA-Z0-9_-]+$")
    description: Optional[str] = None
    parameters: Optional[FunctionParameters] = None


class Tool(BaseModel):
    type: ToolType = ToolType.FUNCTION
    function: Function


class ToolChoice(str, Enum):
    NONE = "none"
    AUTO = "auto"
    REQUIRED = "required"


class SpecificToolChoice(BaseModel):
    type: ToolType = ToolType.FUNCTION
    function: Dict[str, str]


class FunctionCall(BaseModel):
    """Function call details within a tool call."""

    name: str
    arguments: str  # JSON string of arguments


class ToolCall(BaseModel):
    """Tool call from model output."""

    id: str
    type: ToolType = ToolType.FUNCTION
    function: FunctionCall

    @classmethod
    def from_llama_output(
        cls, name: str, parameters: Dict[str, Any], call_id: str
    ) -> "ToolCall":
        """Create a ToolCall instance from Llama model output format."""
        return cls(
            id=call_id,
            type=ToolType.FUNCTION,
            function=FunctionCall(
                name=name,
                arguments=json.dumps(parameters),  # Convert parameters to JSON string
            ),
        )


ToolChoiceType = Union[ToolChoice, SpecificToolChoice]


class Role(str, Enum):
    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"
    TOOL = "tool"


class ChatMessage(BaseModel):
    role: Role
    content: Optional[Union[str, List[Dict[str, str]]]] = None
    name: Optional[str] = None
    tool_calls: Optional[List[ToolCall]] = None
    tool_call_id: Optional[str] = None

    class Config:
        json_encoders = {bytes: lambda v: v.decode()}


class ChatCompletionUsageDetails(BaseModel):
    reasoning_tokens: int = 0
    accepted_prediction_tokens: int = 0
    rejected_prediction_tokens: int = 0


class ChatCompletionUsage(BaseModel):
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class ChatCompletionChoice(BaseModel):
    index: int
    message: ChatMessage
    finish_reason: str
    logprobs: Optional[Dict[str, Any]] = None
    tool_calls: Optional[List[ToolCall]] = None


class ChatCompletionChunkChoice(BaseModel):
    index: int
    delta: ChatMessage
    finish_reason: Optional[str] = None
    logprobs: Optional[Any] = None


class ChatCompletionChunk(BaseModel):
    id: str
    object: str = "chat.completion.chunk"
    created: int
    model: str
    choices: List[ChatCompletionChunkChoice]
    system_fingerprint: Optional[str] = None
    usage: Optional[ChatCompletionUsage] = None


class ChatCompletionResponse(BaseModel):
    id: str
    object: str = "chat.completion"
    created: int
    model: str
    choices: List[ChatCompletionChoice]
    usage: ChatCompletionUsage
    system_fingerprint: Optional[str] = None


class StreamOptions(BaseModel):
    include_usage: bool = False


class JsonSchemaFormat(BaseModel):
    description: Optional[str] = Field(
        None, description="A description of what the response format is for"
    )
    name: str = Field(
        ...,
        description="The name of the response format",
        pattern="^[a-zA-Z0-9_-]{1,64}$",
    )
    schema_def: Optional[Dict[str, Any]] = Field(
        None, description="The schema for the response format", alias="schema"
    )
    strict: Optional[bool] = Field(
        False, description="Whether to enable strict schema adherence"
    )

    @field_validator("name")
    def validate_name(cls, v: str) -> str:
        if len(v) > 64:
            raise ValueError("Name must not exceed 64 characters")
        if not re.match("^[a-zA-Z0-9_-]+$", v):
            raise ValueError(
                "Name must only contain a-z, A-Z, 0-9, underscores and dashes"
            )
        return v


class ResponseFormat(BaseModel):
    type: str = Field(..., description="The type of response format")
    json_schema: Optional[JsonSchemaFormat] = Field(
        None, description="The JSON schema configuration when type is 'json_schema'"
    )

    @field_validator("type")
    def validate_type(cls, v):
        if v not in ["text", "json_object", "json_schema"]:
            raise ValueError(
                "Type must be one of: 'text', 'json_object', or 'json_schema'"
            )
        return v

    @field_validator("json_schema")
    def validate_json_schema(cls, v, values):
        type_val = values.data.get("type")
        if type_val == "json_schema" and v is None:
            raise ValueError("json_schema is required when type is 'json_schema'")
        if type_val != "json_schema" and v is not None:
            raise ValueError(
                "json_schema should only be provided when type is 'json_schema'"
            )
        return v


class ChatCompletionRequest(BaseModel):
    # Standard OpenAI API fields
    model: str = Field(..., description="ID of the model to use")
    messages: List[ChatMessage]
    temperature: Optional[float] = Field(1.0, ge=0, le=2)
    top_p: Optional[float] = Field(1.0, ge=0, le=1)
    max_tokens: Optional[int] = None
    max_completion_tokens: Optional[int] = None
    stream: Optional[bool] = False
    stream_options: Optional[StreamOptions] = None
    seed: Optional[int] = None
    stop: Optional[Union[str, List[str]]] = None
    presence_penalty: Optional[float] = Field(0, ge=-2.0, le=2.0)
    frequency_penalty: Optional[float] = Field(0, ge=-2.0, le=2.0)
    logit_bias: Optional[Dict[str, float]] = None
    logprobs: Optional[bool] = False
    top_logprobs: Optional[int] = Field(
        None,
        ge=0,
        le=20,
    )
    n: Optional[int] = Field(1, ge=1, le=10)
    tools: Optional[List[Tool]] = None
    tool_choice: Optional[ToolChoiceType] = None
    response_format: Optional[ResponseFormat] = None

    # Allow any additional fields
    class Config:
        extra = "allow"  # This allows additional fields not defined in the model

    @field_validator("temperature")
    def validate_temperature(cls, v):
        if v is not None and (v < 0 or v > 2):
            raise ValueError("Temperature must be between 0 and 2")
        return v

    @field_validator("top_p")
    def validate_top_p(cls, v):
        if v is not None and (v < 0 or v > 1):
            raise ValueError("Top_p must be between 0 and 1")
        return v

    def get_extra_params(self) -> Dict[str, Any]:
        """Get all extra parameters that aren't part of the standard OpenAI API."""
        standard_fields: Set[str] = {
            "model",
            "messages",
            "temperature",
            "top_p",
            "max_tokens",
            "max_completion_tokens",
            "stream",
            "seed",
            "stop",
            "presence_penalty",
            "frequency_penalty",
            "logit_bias",
            "logprobs",
            "top_logprobs",
            "n",
            "tools",
            "tool_choice",
            "stream_options",
            "response_format",
        }
        return {k: v for k, v in self.model_dump().items() if k not in standard_fields}



import json
import time
from typing import Callable, Optional

from fastapi import Request, Response
from starlette.middleware.base import BaseHTTPMiddleware

from ..utils.logger import logger


def format_body(body: str) -> str:
    """Format body content for logging."""
    try:
        # Try to parse as JSON and format it
        parsed = json.loads(body)
        return json.dumps(parsed, indent=2, ensure_ascii=False)
    except json.JSONDecodeError:
        # If not JSON, return as is
        return body


class RequestResponseLoggingMiddleware(BaseHTTPMiddleware):
    def __init__(
        self,
        app,
        *,
        exclude_paths: Optional[list[str]] = None,
    ):
        """Initialize the middleware with custom log levels.

        Args:
            app: The ASGI application
            request_level: Logging level for requests (default: INFO)
            response_level: Logging level for responses (default: INFO)
            exclude_paths: List of paths to exclude from logging (default: None)
        """
        super().__init__(app)
        self.exclude_paths = exclude_paths or []

    def should_log(self, path: str) -> bool:
        """Check if the path should be logged."""
        return not any(path.startswith(exclude) for exclude in self.exclude_paths)

    async def dispatch(self, request: Request, call_next: Callable) -> Response:
        if not self.should_log(request.url.path):
            return await call_next(request)

        # Get and parse request body to check if streaming
        body = await self._get_request_body(request)
        try:
            body_json = json.loads(body)
            if body_json.get("stream", False):
                start_time = time.time()
                response = await call_next(request)
                process_time = time.time() - start_time

                logger.debug(
                    f"First Stream Response took {process_time:.2f}s:\n"
                    f"Status: {response.status_code}\n"
                    f"Headers:\n{json.dumps(dict(response.headers), indent=2)}\n"
                )
                return response
        except json.JSONDecodeError:
            pass

        # Generate request ID for tracking
        request_id = str(time.time())

        # Log request
        logger.debug(
            f"Request [{request_id}]: {request.method} {request.url}\n"
            f"Headers:\n{json.dumps(dict(request.headers), indent=2)}\n"
            f"Body:\n{format_body(body)}",
        )

        # Process the request and get response
        start_time = time.time()
        response = await call_next(request)
        process_time = time.time() - start_time

        # For non-streaming responses, continue with normal body logging
        response_body = b""
        async for chunk in response.body_iterator:
            response_body += chunk

        # Create a new async iterator for the response body
        async def body_iterator():
            yield response_body

        response.body_iterator = body_iterator()

        try:
            body_text = response_body.decode()
            body_text = format_body(body_text)
        except UnicodeDecodeError:
            body_text = "<Binary Content>"

        logger.debug(
            f"Response [{request_id}] took {process_time:.2f}s:\n"
            f"Status: {response.status_code}\n"
            f"Headers:\n{json.dumps(dict(response.headers), indent=2)}\n"
            f"Body:\n{body_text}",
        )
        return response

    async def _get_request_body(self, request: Request) -> str:
        """Get request body as string."""
        try:
            body = await request.body()
            return body.decode()
        except Exception:
            return ""



import io

from fastapi import APIRouter, HTTPException
from fastapi.responses import StreamingResponse

from .schema import AudioFormat, TTSRequest
from .tts_service import TTSService

router = APIRouter(tags=["text-to-speech"])


@router.post("/audio/speech")
@router.post("/v1/audio/speech")
async def create_speech(request: TTSRequest):
    """
    Generate audio from input text.

    Returns:
        StreamingResponse: Audio file content in the requested format
    """
    tts_service = TTSService()

    try:
        audio_content = await tts_service.generate_speech(
            request=request,
        )

        # Create content type mapping
        content_type_mapping = {
            AudioFormat.MP3: "audio/mpeg",
            AudioFormat.OPUS: "audio/opus",
            AudioFormat.AAC: "audio/aac",
            AudioFormat.FLAC: "audio/flac",
            AudioFormat.WAV: "audio/wav",
            AudioFormat.PCM: "audio/pcm",
        }

        # Create response
        return StreamingResponse(
            io.BytesIO(audio_content),
            media_type=content_type_mapping[request.response_format],
            headers={
                "Content-Disposition": f'attachment; filename="speech.{request.response_format.value}"'
            },
        )

    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))



import base64
import os
import tempfile
import time
from pathlib import Path
from typing import List

from diffusionkit.mlx import DiffusionPipeline, FluxPipeline

from .schema import ImageGenerationRequest, ImageObject, ResponseFormat


class MLXImageGenerator:
    def __init__(self, model_version: str = "argmaxinc/mlx-FLUX.1-schnell"):
        self.model_version = model_version
        self.height = {
            "argmaxinc/mlx-stable-diffusion-3-medium": 512,
            "argmaxinc/mlx-stable-diffusion-3.5-large": 1024,
            "argmaxinc/mlx-stable-diffusion-3.5-large-4bit-quantized": 1024,
            "argmaxinc/mlx-FLUX.1-schnell": 512,
            "argmaxinc/mlx-FLUX.1-schnell-4bit-quantized": 512,
            "argmaxinc/mlx-FLUX.1-dev": 512,
        }[model_version]

        self.width = {
            "argmaxinc/mlx-stable-diffusion-3-medium": 512,
            "argmaxinc/mlx-stable-diffusion-3.5-large": 1024,
            "argmaxinc/mlx-stable-diffusion-3.5-large-4bit-quantized": 1024,
            "argmaxinc/mlx-FLUX.1-schnell": 512,
            "argmaxinc/mlx-FLUX.1-schnell-4bit-quantized": 512,
            "argmaxinc/mlx-FLUX.1-dev": 512,
        }[model_version]

        self.shift = {
            "argmaxinc/mlx-stable-diffusion-3-medium": 3.0,
            "argmaxinc/mlx-stable-diffusion-3.5-large": 3.0,
            "argmaxinc/mlx-stable-diffusion-3.5-large-4bit-quantized": 3.0,
            "argmaxinc/mlx-FLUX.1-schnell": 1.0,
            "argmaxinc/mlx-FLUX.1-schnell-4bit-quantized": 1.0,
            "argmaxinc/mlx-FLUX.1-dev": 1.0,
        }[model_version]

        self.pipeline_class = (
            FluxPipeline if "FLUX" in model_version else DiffusionPipeline
        )

    def generate(
        self,
        request: ImageGenerationRequest,
        output_path: str,
        low_memory_mode: bool = True,
        **extra_params,
    ):
        """Generate image using MLX DiffusionKit with request parameters"""
        # Extract size parameters from request
        width, height = self._parse_size(request.size)

        # Get extra parameters from request
        request_extra_params = request.get_extra_params()

        # Merge all extra parameters, with passed extra_params taking precedence
        all_extra_params = {**request_extra_params, **extra_params}

        # Initialize model with appropriate settings
        sd = self.pipeline_class(
            w16=True,
            shift=all_extra_params.pop("shift", self.shift),
            use_t5=all_extra_params.pop(
                "use_t5", False
            ),  # Can be overridden by extra_params
            model_version=self.model_version,
            low_memory_mode=low_memory_mode,
            a16=True,
        )

        # Default parameters for generate_image
        gen_params = {
            "text": request.prompt,  # prompt parameter is renamed to text
            "num_steps": all_extra_params.pop("num_steps", 50),
            "cfg_weight": all_extra_params.pop(
                "cfg_weight", 0.0 if "FLUX" in self.model_version else 5.0
            ),
            "negative_text": all_extra_params.pop("negative_prompt", ""),
            "latent_size": all_extra_params.pop(
                "latent_size", (height // 8, width // 8)
            ),
            "seed": all_extra_params.pop("seed", None),
            "verbose": all_extra_params.pop("verbose", False),
            "image_path": all_extra_params.pop("image_path", None),
            "denoise": all_extra_params.pop("denoise", 1.0),
        }

        # Add any remaining extra parameters
        gen_params.update(all_extra_params)

        # Generate image
        image, _ = sd.generate_image(**gen_params)

        # Save generated image
        image.save(output_path)
        return image

    def _parse_size(self, size_str: str) -> tuple[int, int]:
        """Parse size string into width and height"""
        try:
            width, height = map(int, size_str.split("x"))
            return width, height
        except (ValueError, AttributeError):
            # Default to model's default size if parsing fails
            return self.width, self.height


class ImagesService:
    def __init__(self):
        # Use system temporary directory
        self.output_dir = Path(tempfile.gettempdir()) / "mlx_omni_server" / "images"
        self.output_dir.mkdir(parents=True, exist_ok=True)

    def _get_output_path(self, uid: str) -> str:
        """Generate unique output path for image"""
        return str(self.output_dir / f"{uid}.png")

    def _image_to_base64(self, image_path: str) -> str:
        """Convert image to base64 string"""
        with open(image_path, "rb") as img_file:
            return base64.b64encode(img_file.read()).decode("utf-8")

    def _cleanup_image(self, image_path: str):
        """Clean up temporary image file"""
        try:
            os.unlink(image_path)
        except Exception as e:
            print(f"Error cleaning up image {image_path}: {str(e)}")

    def generate_images(
        self,
        request: ImageGenerationRequest,
    ) -> List[ImageObject]:
        """Generate images based on the request"""
        generated_images = []
        generator = MLXImageGenerator(model_version=request.model)

        for i in range(request.n):
            # Generate unique identifier for this image
            uid = f"{int(time.time())}_{i}"
            output_path = self._get_output_path(uid)

            try:
                # Generate the image
                generator.generate(
                    request=request, output_path=output_path, low_memory_mode=True
                )

                # Create response object based on format
                image_object = ImageObject(revised_prompt=request.prompt)

                # Response All Format
                image_object.b64_json = self._image_to_base64(output_path)
                image_object.url = f"file://{output_path}"

                generated_images.append(image_object)

            except Exception as e:
                raise Exception(f"Error generating image: {str(e)}")
            finally:
                # Clean up temporary file if using base64 format
                if request.response_format == ResponseFormat.B64_JSON:
                    self._cleanup_image(output_path)

        return generated_images
